use std::{
    cmp::PartialEq,
    ops::{AddAssign, BitAndAssign, BitOrAssign, BitXorAssign, DivAssign, MulAssign, SubAssign},
};

use crate::utils::ulps_eq;
use crate::{difference_of_products, two_prod, two_sum};

//-------------------------------------------------------------
// Which implementation to use.

#[cfg(not(any(target_arch = "x86_64")))]
mod fallback;
#[cfg(not(any(target_arch = "x86_64")))]
pub use fallback::{Bool4, Float4};

#[cfg(target_arch = "x86_64")]
mod sse;
#[cfg(target_arch = "x86_64")]
pub use sse::{Bool4, Float4};

//-------------------------------------------------------------

impl Float4 {
    /// 3D dot product (only uses the first 3 components).
    #[inline(always)]
    pub fn dot_3(a: Self, b: Self) -> f32 {
        let (p, p_err) = two_prod(a, b);

        // Products.
        let (x, x_err) = (p.a(), p_err.a());
        let (y, y_err) = (p.b(), p_err.b());
        let (z, z_err) = (p.c(), p_err.c());

        // Sums.
        let (s1, s1_err) = two_sum(x, y);
        let err1 = x_err + (y_err + s1_err);

        let (s2, s2_err) = two_sum(s1, z);
        let err2 = z_err + (err1 + s2_err);

        // Final result with rounding error compensation.
        s2 + err2
    }

    /// 3D dot product (only uses the first 3 components).
    ///
    /// Faster but less precise version.
    #[inline(always)]
    pub fn dot_3_fast(a: Self, b: Self) -> f32 {
        let c = a * b;
        c.a() + c.b() + c.c()
    }

    /// 3D cross product (only uses the first 3 components).
    #[inline(always)]
    pub fn cross_3(a: Self, b: Self) -> Self {
        difference_of_products(a.bcad(), b.cabd(), a.cabd(), b.bcad())
    }

    /// 3D cross product (only uses the first 3 components).
    ///
    /// Faster but less precise version.
    #[inline(always)]
    pub fn cross_3_fast(a: Self, b: Self) -> Self {
        (a.bcad() * b.cabd()) - (a.cabd() * b.bcad())
    }

    #[inline(always)]
    pub fn transpose_3x3(m: [Self; 3]) -> [Self; 3] {
        [
            // The fourth component in each row below is arbitrary,
            // but in this case chosen so that it matches the
            // behavior of the SSE version of transpose_3x3.
            Self::new(m[0].a(), m[1].a(), m[2].a(), m[2].d()),
            Self::new(m[0].b(), m[1].b(), m[2].b(), m[2].d()),
            Self::new(m[0].c(), m[1].c(), m[2].c(), m[2].d()),
        ]
    }

    /// Invert a 3x3 matrix.
    ///
    /// Returns `None` if not invertible.
    #[inline]
    pub fn invert_3x3(m: [Self; 3]) -> Option<[Self; 3]> {
        let m0_bca = m[0].bcad();
        let m1_bca = m[1].bcad();
        let m2_bca = m[2].bcad();
        let m0_cab = m[0].cabd();
        let m1_cab = m[1].cabd();
        let m2_cab = m[2].cabd();
        let abc = difference_of_products(m1_bca, m2_cab, m1_cab, m2_bca);
        let def = difference_of_products(m2_bca, m0_cab, m2_cab, m0_bca);
        let ghi = difference_of_products(m0_bca, m1_cab, m0_cab, m1_bca);

        let det = Self::dot_3(
            Self::new(abc.a(), def.a(), ghi.a(), 0.0),
            Self::new(m[0].a(), m[1].a(), m[2].a(), 0.0),
        );

        if det == 0.0 {
            None
        } else {
            Some(Self::transpose_3x3([abc / det, def / det, ghi / det]))
        }
    }

    /// Invert a 3x3 matrix.  Faster but less precise version.
    ///
    /// Returns `None` if not invertible.
    #[inline]
    pub fn invert_3x3_fast(m: [Self; 3]) -> Option<[Self; 3]> {
        let m0_bca = m[0].bcad();
        let m1_bca = m[1].bcad();
        let m2_bca = m[2].bcad();
        let m0_cab = m[0].cabd();
        let m1_cab = m[1].cabd();
        let m2_cab = m[2].cabd();
        let abc = (m1_bca * m2_cab) - (m1_cab * m2_bca);
        let def = (m2_bca * m0_cab) - (m2_cab * m0_bca);
        let ghi = (m0_bca * m1_cab) - (m0_cab * m1_bca);

        let det = Self::dot_3_fast(
            Self::new(abc.a(), def.a(), ghi.a(), 0.0),
            Self::new(m[0].a(), m[1].a(), m[2].a(), 0.0),
        );

        if det == 0.0 {
            None
        } else {
            Some(Self::transpose_3x3([abc / det, def / det, ghi / det]))
        }
    }

    /// Multiplies a 3D vector with a 3x3 matrix.
    #[inline]
    pub fn vec_mul_3x3(self, m: &[Self; 3]) -> Self {
        let x = self.aaaa();
        let y = self.bbbb();
        let z = self.cccc();

        // Products.
        let (a, a_err) = two_prod(x, m[0]);
        let (b, b_err) = two_prod(y, m[1]);
        let (c, c_err) = two_prod(z, m[2]);

        // Sums.
        let (s1, s1_err) = two_sum(a, b);
        let err1 = a_err + (b_err + s1_err);

        let (s2, s2_err) = two_sum(c, s1);
        let err2 = c_err + (err1 + s2_err);

        s2 + err2
    }

    /// Multiplies a 3D vector with a 3x3 matrix.
    ///
    /// Faster but less precise version.
    #[inline]
    pub fn vec_mul_3x3_fast(self, m: &[Self; 3]) -> Self {
        let x = self.aaaa();
        let y = self.bbbb();
        let z = self.cccc();

        (x * m[0]) + (y * m[1]) + (z * m[2])
    }

    /// Transforms a 3d point by an affine transform.
    ///
    /// `m` is the 3x3 part of the affine transform, `t` is the translation part.
    #[inline]
    pub fn vec_mul_affine(self, m: &[Self; 3], t: Self) -> Self {
        let x = self.aaaa();
        let y = self.bbbb();
        let z = self.cccc();

        // Products.
        let (a, a_err) = two_prod(x, m[0]);
        let (b, b_err) = two_prod(y, m[1]);
        let (c, c_err) = two_prod(z, m[2]);

        // Sums.
        let (s1, s1_err) = two_sum(a, b);
        let err1 = a_err + (b_err + s1_err);

        let (s2, s2_err) = two_sum(c, s1);
        let err2 = c_err + (err1 + s2_err);

        let (s3, s3_err) = two_sum(t, s2);
        let err3 = err2 + s3_err;

        s3 + err3
    }

    /// Transforms a 3d point by an affine transform.
    ///
    /// Faster but less precise version.
    #[inline]
    pub fn vec_mul_affine_fast(self, m: &[Self; 3], t: Self) -> Self {
        let x = self.aaaa();
        let y = self.bbbb();
        let z = self.cccc();

        (x * m[0]) + (y * m[1]) + (z * m[2]) + t
    }

    /// Transforms a 3d point by an affine transform, except it applies
    /// the translation part before the 3x3 part.
    ///
    /// This is primarily useful for performing efficient inverse transforms by
    /// passing an inverted 3x3 part and a negated translation part.
    ///
    /// `m` is the 3x3 part of the affine transform, `t` is the translation part.
    #[inline]
    pub fn vec_mul_affine_rev(self, m: &[Self; 3], t: Self) -> Self {
        let (v, v_err) = two_sum(self, t);

        let (x, x_err) = (v.aaaa(), v_err.aaaa());
        let (y, y_err) = (v.bbbb(), v_err.bbbb());
        let (z, z_err) = (v.cccc(), v_err.cccc());

        // Products.
        let ((a, a_err1), a_err2) = (two_prod(x, m[0]), x_err * m[0]);
        let ((b, b_err1), b_err2) = (two_prod(y, m[1]), y_err * m[1]);
        let ((c, c_err1), c_err2) = (two_prod(z, m[2]), z_err * m[2]);
        let a_err = a_err1 + a_err2;
        let b_err = b_err1 + b_err2;
        let c_err = c_err1 + c_err2;

        // Sums.
        let (s1, s1_err) = two_sum(a, b);
        let err1 = a_err + (b_err + s1_err);

        let (s2, s2_err) = two_sum(c, s1);
        let err2 = c_err + (err1 + s2_err);

        s2 + err2
    }

    /// Transforms a 3d point by an affine transform, except it applies
    /// the translation part before the 3x3 part.
    ///
    /// Faster but less precise version.
    #[inline]
    pub fn vec_mul_affine_rev_fast(self, m: &[Self; 3], t: Self) -> Self {
        let v = self + t;

        let x = v.aaaa();
        let y = v.bbbb();
        let z = v.cccc();

        (x * m[0]) + (y * m[1]) + (z * m[2])
    }

    /// Returns whether the `Float4`s are approximately equal to each
    /// other.
    ///
    /// Each corresponding element cannot have a relative error exceeding
    /// `epsilon`.
    pub(crate) fn aprx_eq(a: Self, b: Self, max_ulps: u32) -> bool {
        let mut eq = true;
        eq &= ulps_eq(a.a(), b.a(), max_ulps);
        eq &= ulps_eq(a.b(), b.b(), max_ulps);
        eq &= ulps_eq(a.c(), b.c(), max_ulps);
        eq &= ulps_eq(a.d(), b.d(), max_ulps);
        eq
    }

    /// Transforms one affine transform by another.
    ///
    /// The result is an affine transform that acts as a sequence of the
    /// first followed by the second.
    ///
    /// `m#` is the 3x3 part of the affine transform, `t#` is the translation part.
    #[inline]
    pub fn affine_mul_affine(
        m1: &[Self; 3],
        t1: Self,
        m2: &[Self; 3],
        t2: Self,
    ) -> ([Self; 3], Self) {
        (
            [
                m1[0].vec_mul_3x3(&m2),
                m1[1].vec_mul_3x3(&m2),
                m1[2].vec_mul_3x3(&m2),
            ],
            t1.vec_mul_affine(&m2, t2),
        )
    }

    /// Transforms one affine transform by another.
    ///
    /// Faster but less precise version.
    #[inline]
    pub fn affine_mul_affine_fast(
        m1: &[Self; 3],
        t1: Self,
        m2: &[Self; 3],
        t2: Self,
    ) -> ([Self; 3], Self) {
        (
            [
                m1[0].vec_mul_3x3_fast(&m2),
                m1[1].vec_mul_3x3_fast(&m2),
                m1[2].vec_mul_3x3_fast(&m2),
            ],
            t1.vec_mul_affine_fast(&m2, t2),
        )
    }
}

impl From<Float4> for (f32, f32, f32, f32) {
    fn from(v: Float4) -> (f32, f32, f32, f32) {
        (v.a(), v.b(), v.c(), v.d())
    }
}

impl AddAssign for Float4 {
    #[inline(always)]
    fn add_assign(&mut self, rhs: Self) {
        *self = *self + rhs;
    }
}

impl SubAssign for Float4 {
    #[inline(always)]
    fn sub_assign(&mut self, rhs: Self) {
        *self = *self - rhs;
    }
}

impl MulAssign for Float4 {
    #[inline(always)]
    fn mul_assign(&mut self, rhs: Self) {
        *self = *self * rhs;
    }
}

impl MulAssign<f32> for Float4 {
    #[inline(always)]
    fn mul_assign(&mut self, rhs: f32) {
        *self = *self * rhs;
    }
}

impl DivAssign for Float4 {
    #[inline(always)]
    fn div_assign(&mut self, rhs: Self) {
        *self = *self / rhs;
    }
}

impl DivAssign<f32> for Float4 {
    #[inline(always)]
    fn div_assign(&mut self, rhs: f32) {
        *self = *self / rhs;
    }
}

impl PartialEq for Float4 {
    #[inline(always)]
    fn eq(&self, rhs: &Self) -> bool {
        self.cmpeq(*rhs).bitmask() == 0b1111
    }
}

//--------

impl BitAndAssign for Bool4 {
    #[inline(always)]
    fn bitand_assign(&mut self, rhs: Self) {
        *self = *self & rhs;
    }
}

impl BitOrAssign for Bool4 {
    #[inline(always)]
    fn bitor_assign(&mut self, rhs: Self) {
        *self = *self | rhs;
    }
}

impl BitXorAssign for Bool4 {
    #[inline(always)]
    fn bitxor_assign(&mut self, rhs: Self) {
        *self = *self ^ rhs;
    }
}

//-------------------------------------------------------------

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn approximate_equality_test() {
        let a = Float4::new(1.0, 2.0, 3.0, 4.0);
        let b = Float4::new(1.00001, 2.00002, 3.00003, 4.00004);
        let c = Float4::new(1.0e-43, 2.0e-43, 3.0e-43, 4.0e-43);
        let d = Float4::new(-1.0e-43, -2.0e-43, -3.0e-43, -4.0e-43);

        assert!(Float4::aprx_eq(a, a, 0));

        assert!(Float4::aprx_eq(a, b, 130));
        assert!(!Float4::aprx_eq(a, b, 120));

        assert!(Float4::aprx_eq(c, d, 575));
        assert!(!Float4::aprx_eq(c, d, 565));
    }

    #[test]
    fn index() {
        let v = Float4::new(0.0, 1.0, 2.0, 3.0);
        assert_eq!(v[0], 0.0);
        assert_eq!(v[1], 1.0);
        assert_eq!(v[2], 2.0);
        assert_eq!(v[3], 3.0);
    }

    #[test]
    fn shuffle() {
        let v = Float4::new(0.0, 1.0, 2.0, 3.0);

        assert_eq!(v.aaaa(), Float4::splat(0.0));
        assert_eq!(v.bbbb(), Float4::splat(1.0));
        assert_eq!(v.cccc(), Float4::splat(2.0));
        assert_eq!(v.dddd(), Float4::splat(3.0));

        assert_eq!(v.bcad(), Float4::new(1.0, 2.0, 0.0, 3.0));
        assert_eq!(v.cabd(), Float4::new(2.0, 0.0, 1.0, 3.0));
    }

    #[test]
    fn bitmask() {
        let v1 = Float4::new(0.0, 1.0, 2.0, 3.0);
        let v2 = Float4::new(9.0, 1.0, 9.0, 3.0);

        assert_eq!(v1.cmpeq(v2).bitmask(), 0b1010);
    }
}
